#!/bin/env python

import struct

from optparse import OptionParser


def hexdump(src, length=16, sep='.'):
    FILTER = ''.join([(len(repr(chr(x))) == 3) and chr(x) or sep for x in range(256)])
    lines = []
    for c in xrange(0, len(src), length):
        chars = src[c:c+length]
	hex = ' '.join(["%02x" % ord(x) for x in chars])
	if len(hex) > 24:
	    hex = "%s %s" % (hex[:24], hex[24:])
        printable = ''.join(["%s" % ((ord(x) <= 127 and FILTER[ord(x)]) or sep) for x in chars])
        lines.append("%08x:  %-*s  |%s|\n" % (c, length*3, hex, printable))
    return ''.join(lines)


class SourceFile(object):
    def __init__(self, fname):
        self.fname = fname
        self.lineno = 0
        self.line = ''
        self.fh = open(fname)

    def token(self, sep):
        global options
        while self.line == '':
            self.lineno += 1
            line = self.fh.readline()
            if len(line) == 0:
                self.fh.close()
                return None
            line = line.expandtabs()
            self.line = line.rstrip()
            self.full_line = self.line
            if options.debug > 1:
                print self.full_line
        line = self.line.lstrip(sep)
        if sep in line:
            tok, rest = line.split(sep, 1)
            self.line = rest
        else:
            tok = line
            self.line = ''
        return tok

    def error(self, s):
        global options
        if options.debug:
            print self.full_line
        raise RuntimeError("%s:%d: %s" % (self.fname, self.lineno, s))

class Stack(object):
    def __init__(self):
        self.s = []

    def push(self, el):
        self.s.append(el)

    def pop(self):
        if len(self.s) == 0:
            raise RuntimeError("Stack underflow")
        return self.s.pop(-1)

    def peek(self):
        if len(self.s) == 0:
            return 0
        return self.s[-1]

    def __repr__(self):
        s = "<%d> " % len(self.s)
        s += ' '.join(map(lambda x: "%x" % x, self.s))
        return s

class TargetBlock(object):
    def __init__(self, start):
        self.start = start
        self.mem = ''

    def valid(self, addr):
        if addr < self.start:
            return None
        if addr-self.start > len(self.mem):
            return None
        return self

    def here(self):
        return self.start + len(self.mem)

    def fetch(self, addr):
        if not self.valid(addr):
            raise IndexError, "Address 0x%04x out of range" % addr
        o = addr-self.start
        return struct.unpack('<H', self.mem[o:o+2])[0]

    def cfetch(self, addr):
        if not self.valid(addr):
            raise IndexError, "Address 0x%04x out of range" % addr
        o = addr-self.start
        return struct.unpack('<B', self.mem[o])[0]

    def store(self, addr, w):
        if not self.valid(addr):
            raise IndexError, "Address 0x%04x out of range" % addr
        o = addr-self.start
        fmt = w < 0 and '<h' or '<H'
        self.mem = self.mem[:o]+struct.pack(fmt, w)+self.mem[o+2:]

    def cstore(self, addr, b):
        if not self.valid(addr):
            raise IndexError, "Address 0x%04x out of range" % addr
        o = addr-self.start
        self.mem = self.mem[:o]+struct.pack('<B', b)+self.mem[o+1:]

    def comma(self, w):
        fmt = w < 0 and '<h' or '<H'
        self.mem += struct.pack(fmt, w)

    def ccomma(self, b):
        self.mem += struct.pack('<B', b)

    def append(self, s):
        self.mem += s

    def last_setb(self, b):
	self.mem = self.mem[:-1]+struct.pack('<B', ord(self.mem[-1:]) | b)

    def __repr__(self):
        return hexdump(self.mem)

class TargetDict(object):
    def __init__(self, start, aux_start):
        self.main_block = TargetBlock(start)
        if aux_start:
            # ROM-based target
            self.aux_block = TargetBlock(aux_start)
        else:
            self.aux_block = None
        self.block = self.main_block
        self.mem = ''
        self.latest = 0
        self.fwd_refs = {}
        self.dict_head = 0
        self.voc_link = 0

    def get_block(self, addr):
        if self.aux_block and self.aux_block.valid(addr):
            return self.aux_block.valid(addr)
        if not self.main_block.valid(addr):
            raise RuntimeError("Invalid memory access @%04x" % addr)
        return self.main_block

    def here(self):
        return self.block.here()

    def fetch(self, addr):
        return self.get_block(addr).fetch(addr)

    def cfetch(self, addr):
        return self.get_block(addr).cfetch(addr)

    def store(self, addr, w):
        return self.get_block(addr).store(addr, w)

    def cstore(self, addr, b):
        return self.get_block(addr).cstore(addr, b)

    def comma(self, w):
        self.block.comma(w)

    def ccomma(self, b):
        self.block.ccomma(b)

    def append(self, s):
        self.block.append(s)

    def create(self, w, d, cfa = None):
        global options
        w = w.upper()
        new = self.here()
        self.ccomma(len(w) | 0x80)
        self.append(w[:min(0x1F, len(w))])
        if (self.here() & 0xff) == 0xfd:
            if options.debug:
                print "CFA of '%s' on page boundary %04x, padding" % (w, self.here()+2)
            self.ccomma(0x20)
        self.block.last_setb(0x80)
        self.comma(d.latest) # LFA
        if not cfa:
            # CODE word, machine code follows
            cfa = self.here()+2
        self.comma(cfa) # CFA
	self.latest = new
	d.latest = new
        if self.dict_head:
            self.store(self.dict_head, new)

    def smudge(self):
        self.cstore(self.latest, self.cfetch(self.latest) ^ 0x20)

    def immediate(self):
        self.cstore(self.latest, self.cfetch(self.latest) | 0x40)

    def traverse(self, nfa):
        global options
        a = nfa + 1
        while not self.cfetch(a) & 0x80:
            a += 1
        a += 1
        if a & 0xff == 0xfd:
            if options.debug:
                print "Adjusting CFA of %04x to skip page boundary" % nfa
            a += 1
        return a

    def tick(self, w, missing_ok = False):
        global dicts
        if w.nfa != 0:
            return self.traverse(w.nfa)+2
        if not missing_ok:
            raise RuntimeError("No CFA for "+w.name)
        self.fwd_ref(dicts.current, dicts.context, w.name)
        return 0;

    def body(self, w):
        if w.nfa != 0:
            return self.traverse(w.nfa)+4
        raise RuntimeError("No PFA for "+w.name)

    def fwd_ref(self, current, context, name):
        global options
        if options.debug > 1:
            print "Forward reference to '%s' @%04x" % (name, self.here())
        ref = (current, context, name)
        if ref not in self.fwd_refs:
            self.fwd_refs[ref] = []
        self.fwd_refs[ref].append(self.here())

class Word(object):
    def __init__(self, name, fn = None, args = (), runtime_fn = None,
                 value = None, immediate = False):
        self.name = name
        self.fn = fn
        self.args = args
        self.runtime_fn = runtime_fn
        self.value = value
        self.immediate = immediate
        self.nfa = 0
        self.defn = None
        self.runtime_cfa = None

    def execute(self):
        global stack
        global xstack
        global options
        xstack.push(self)
        if options.debug > 1:
            print "Executing "+self.name
        if self.fn:
            self.fn(self, *self.args)
        elif isinstance(self.defn, list):
            for w in self.defn:
                if isinstance(w, Word):
                    w.execute()
                elif isinstance(w, int):
                    stack.push(w)
                else:
                    raise RuntimeError("Unknown definition, "+w)
        else:
            raise RuntimeError("No implementation for %s" % self.name)
        xstack.pop()

class CrossDict(object):
    def __init__(self, parent):
        self.parent = parent
        self.words = {}
        self.latest = 0

    def add(self, word):
        if word.name in self.words:
            print "Warning: '%s' redefined" % word.name
        self.words[word.name] = word

    def constant(self, name, val):
        self.add(Word(name, fn = inline_constant, value = val, immediate = True))

    def find(self, name):
        d = self.words.get(name)
        if not d and self.parent:
            d = self.parent.words.get(name)
        return d

class DictChain(dict):
    def __init__(self):
        self.current = None # dictionary to add to
        self.context = None # dictionary to search
        self.latest = None
        self.dicts = {}

    def create(self, name):
        exist = self.dicts[self.current].find(name)
        if exist:
            return exist
        w = Word(name)
        self.dicts[self.current].add(w)
        return w

def get_word():
    global src
    return src.token(' ').lower()

def find(tok):
    global dicts
    global options
    w = dicts.dicts[dicts.context].find(tok)
    if not w:
        w = dicts.dicts[dicts.current].find(tok)
    if options.debug > 2:
        print "    dict lookup ", tok, w
    return w

def tick(name, missing_ok = False):
    global target
    global dicts
    w = find(name)
    if w:
        return target.tick(w, missing_ok)
    if not missing_ok:
        raise RuntimeError("No CFA for "+name)
    target.fwd_ref(dicts.current, dicts.context, name)
    return 0;

def body(name):
    global target
    w = find(name)
    if w:
        return target.body(w)
    raise RuntimeError("No PFA for "+name)

def value_get(var):
    w = find(var)
    if not w:
        global src
        src.error("Unknown variable, "+var)
    return w.value

def value_set(var, val):
    w = find(var)
    if not w:
        global src
        src.error("Unknown variable, "+var)
    w.value = val

def number(tok):
    global dicts
    global stack
    base = value_get('base')
    try:
        if tok.endswith('.'):
            # double
            return long(tok.rstrip('.'), base)
        else:
            return int(tok, base)
    except:
        pass
    return None

def interpret(tok):
    global dicts
    global stack
    global target
    state = value_get('state')
    w = find(tok)
    if state == 0:
        # interpret
        if w:
            w.execute()
        else:
            n = number(tok)
            if isinstance(n, int):
                stack.push(n)
            elif isinstance(n, long):
                stack.push(n & 0xffff)
                stack.push((n >> 16) & 0xffff)
            else:
                global src
                src.error("Cannot interpret '%s', undefined" % tok)
    else:
        # compiling
        if w:
            if w.immediate:
                w.execute()
            else:
                target.comma(target.tick(w, missing_ok = True))
                dicts.latest.defn.append(w)
        else:
            n = number(tok)
            if isinstance(n, int):
                if n >= 0 and n <= 0xff:
                    target.comma(tick('clit'))
                    target.ccomma(n)
                else:
                    target.comma(tick('lit'))
                    target.comma(n)
                dicts.latest.defn.append(n)
            elif isinstance(n, long):
                n1 = int(n & 0xffff)
                n2 = int((n >> 16) & 0xffff)
                target.comma(tick('lit'))
                target.comma(n1)
                target.comma(tick('lit'))
                target.comma(n2)
                dicts.latest.defn.append(n1)
                dicts.latest.defn.append(n2)
            else:
                # Undefined, forward reference
                target.comma(tick(tok, missing_ok = True))
                dicts.latest.defn.append(tok)

########################################
# FORTH dictionary words

def left_bracket(w):
    value_set('state', 0)

def right_bracket(w):
    value_set('state', 0xC0)

def here(w):
    global target
    global stack
    stack.push(target.here())

def aux_here(w):
    global target
    global stack
    if not target.aux_block:
        raise RuntimeError("No auxiliary block")
    stack.push(target.aux_block.here())

def comment(w):
    global src
    global target
    comment = src.token(')').strip()
    if target.aux_block:
        if comment == '{mutable}':
            target.block = target.aux_block
        elif comment == '{immutable}':
            target.block = target.main_block

def cstore(w):
    global stack
    global target
    addr = stack.pop()
    val = stack.pop()
    target.cstore(addr, val)

def store(w):
    global stack
    global target
    addr = stack.pop()
    val = stack.pop()
    target.store(addr, val)

def fetch(w):
    global stack
    global target
    addr = stack.pop()
    stack.push(target.fetch(addr))

def ccomma(w):
    global target
    global stack
    target.ccomma(stack.pop())

def comma(w):
    global target
    global stack
    target.comma(stack.pop())

def decimal(w):
    value_set('base', 10)

def forth_hex(w):
    value_set('base', 16)

def run_constant(w):
    global stack
    global target
    if w.value is not None:
        stack.push(w.value)
    else:
        stack.push(target.fetch(target.body(w)))

def inline_constant(w):
    global stack
    global target
    state = value_get('state')
    if state == 0:
        stack.push(w.value)
    else:
        target.comma(tick('lit'))
        target.comma(w.value)

def run_variable(w):
    global stack
    global target
    if w.value is None:
        stack.push(target.body(w))
    else:
        stack.push(w.value)

def run_user(w):
    global stack
    global target
    if w.value is None:
        stack.push(target.block.start+12+target.fetch(target.body(w)))
    else:
        stack.push(w.value)

def vocabulary(w):
    global dicts
    global target
    current = dicts.dicts[dicts.current]
    d = get_word()
    new_w = dicts.create(d)
    new_w.fn = w.runtime_fn
    here = target.here()
    target.create(d, current, w.runtime_cfa)
    new_w.nfa = here
    target.comma(0xA081)
    target.dict_head = target.here()
    target.comma(0)
    voc_link = target.here()
    target.comma(target.voc_link)
    target.voc_link = voc_link
    if d not in dicts.dicts:
        new_d = CrossDict(current)
        new_d.latest = target.tick(current.find(dicts.current))+2
        dicts.dicts[d] = new_d

def dovoc(w):
    global dicts
    dicts.context = w.name

def definitions(w):
    global dicts
    dicts.current = dicts.context

def create(w):
    global dicts
    global target
    global xstack
    global options
    defn_w = xstack.s[0]
    d = get_word()
    if options.debug > 1:
        print "Creating '%s' as a %s in %s" % (d, defn_w.name, dicts.current)
    new_w = dicts.create(d)
    fn = defn_w.runtime_fn
    if fn is None:
        fn = w.runtime_fn
    new_w.fn = fn
    here = target.here()
    cfa = defn_w.runtime_cfa
    if cfa is None:
        cfa = w.runtime_cfa
    target.create(d, dicts.dicts[dicts.current], cfa)
    new_w.nfa = here

def docreate(w):
    global target
    global stack
    stack.push(target.body(w))

def code(w):
    global dicts
    global target
    d = get_word()
    new_w = dicts.create(d)
    new_w.nfa = target.here()
    target.create(d, dicts.dicts[dicts.current])
    target.smudge()
    dicts.context = 'assembler'

def colon(w):
    global dicts
    global target
    d = get_word()
    new_w = dicts.create(d)
    here = target.here()
    target.create(d, dicts.dicts[dicts.current], w.runtime_cfa)
    new_w.nfa = here
    new_w.defn = []
    dicts.context = dicts.current
    dicts.latest = new_w
    target.smudge()
    value_set('state', 0xC0)

def semi(w):
    target.comma(tick('exit'))
    target.smudge()
    value_set('state', 0)

def semi_code(w):
    global dicts
    global target
    global options
    target.comma(tick('(;code)', missing_ok = True))
    if options.debug:
        print "Setting runtime CFA of '%s' to %04x" % (dicts.latest.name, target.here())
    dicts.latest.runtime_cfa = target.here()
    if dicts.latest.name == ':':
        target.store(tick(':'), target.here())
    value_set('state', 0)
    dicts.context = 'assembler'

def does(w):
    global dicts
    global target
    global options
    target.comma(tick('(;code)'))
    if options.debug:
        print "Setting runtime CFA of '%s' to %04x" % (dicts.latest.name, target.here())
    dicts.latest.runtime_cfa = target.here()
    target.ccomma(0x4C)
    target.comma(body('dodoes'))

def forth_tick(w):
    global stack
    d = get_word()
    stack.push(tick(d))

def immediate(w):
    global target
    target.immediate()

def plus_origin(w):
    global target
    global stack
    off = stack.pop()
    stack.push(target.block.start+off)

def aux_origin(w):
    global target
    global stack
    if not target.aux_block:
        raise RuntimeError("No auxiliary block")
    stack.push(target.aux_block.start)

def br_compile(w):
    global target
    name = get_word()
    target.comma(tick(name, missing_ok = True))

def literal(w):
    global target
    global stack
    state = value_get('state')
    if state != 0:
        target.comma(tick('lit'))
	target.comma(stack.pop())

def allot(w):
    global target
    global stack
    target.append('\0' * stack.pop())

def latest(w):
    global target
    global stack
    stack.push(target.latest)

def dot_quote(w):
    global target
    global src
    s = src.token('"')
    target.comma(tick('(.")'))
    target.ccomma(len(s))
    target.append(s)

def one_plus(w):
    global stack
    stack.push(stack.pop()+1)

def two_plus(w):
    global stack
    stack.push(stack.pop()+2)

def two_mult(w):
    global stack
    stack.push(stack.pop()*2)

def minus(w):
    global stack
    a = stack.pop()
    b = stack.pop()
    stack.push(b-a)

def plus(w):
    global stack
    a = stack.pop()
    b = stack.pop()
    stack.push(b+a)

def div_mod(w):
    global stack
    a = stack.pop()
    b = stack.pop()
    stack.push(b%a)
    stack.push(b/a)

def swap(w):
    global stack
    t = stack.s[-1]
    stack.s[-1] = stack.s[-2]
    stack.s[-2] = t

def rot(w):
    global stack
    t = stack.s[-3]
    stack.s[-3] = stack.s[-2]
    stack.s[-2] = stack.s[-1]
    stack.s[-1] = t

def forth_if(w):
    global target
    global stack
    target.comma(tick('0branch'))
    stack.push(target.here())
    target.comma(0)
    stack.push(2)

def forth_then(w):
    global target
    global stack
    global src
    state = value_get('state')
    if state == 0:
        src.error("Conditional outside compilation")
    if stack.pop() != 2:
        src.error("Mismatched THEN")
    fwd = stack.pop()
    target.store(fwd, target.here()-fwd)

def forth_else(w):
    global target
    global stack
    if stack.pop() != 2:
        global src
        src.error("Mismatched ELSE")
    target.comma(tick('branch'))
    stack.push(target.here())
    target.comma(0)
    swap(w)
    stack.push(2)
    forth_then(w)
    stack.push(2)

def forth_do(w):
    global target
    global stack
    target.comma(tick('(do)'))
    stack.push(target.here())
    target.comma(0)
    stack.push(3)

def query_do(w):
    global target
    global stack
    target.comma(tick('(?do)'))
    stack.push(target.here())
    target.comma(0)
    stack.push(3)

def loop(w):
    global target
    global stack
    if stack.pop() != 3:
        global src
        src.error("Mismatched LOOP")
    target.comma(tick('(loop)'))
    doaddr = stack.pop()
    target.store(doaddr, target.here())

def plus_loop(w):
    global target
    global stack
    if stack.pop() != 3:
        global src
        src.error("Mismatched +LOOP")
    target.comma(tick('(+loop)'))
    doaddr = stack.pop()
    target.store(doaddr, target.here())

def begin(w):
    global target
    global stack
    state = value_get('state')
    if state == 0:
        global src
        src.error("Conditional outside compilation")
    stack.push(target.here())
    stack.push(1)

def until(w):
    global target
    global stack
    if stack.pop() != 1:
        global src
        src.error("Mismatched UNTIL")
    target.comma(tick('0branch'))
    target.comma(stack.pop()-target.here())

def again(w):
    global target
    global stack
    if stack.pop() != 1:
        global src
        src.error("Mismatched AGAIN")
    target.comma(tick('branch'))
    target.comma(stack.pop()-target.here())

def repeat(w):
    global stack
    v1 = stack.pop()
    v2 = stack.pop()
    again(w)
    stack.push(v2)
    stack.push(v1-2)
    forth_then(w)

def forth_while(w):
    global stack
    forth_if(w)
    stack.push(stack.pop()+2)

def to_r(w):
    global rstack
    rstack.push(stack.pop())

def r_fetch(w):
    global rstack
    stack.push(rstack.peek())

def r_from(w):
    global rstack
    stack.push(rstack.pop())

def setup_forth_dict():
    global dicts
    forth = CrossDict(None)
    forth.add(Word('[', fn = left_bracket, immediate = True))
    forth.add(Word(']', fn = right_bracket))
    forth.add(Word('here', fn = here))
    forth.add(Word('aux.here', fn = aux_here))
    forth.add(Word('(', fn = comment, immediate = True))
    forth.add(Word('c!', fn = cstore))
    forth.add(Word('!', fn = store))
    forth.add(Word('@', fn = fetch))
    forth.add(Word('c,', fn = ccomma))
    forth.add(Word(',', fn = comma))
    forth.add(Word('base', fn = run_variable, value = 10))
    forth.add(Word('state', fn = run_variable, value = 0))
    forth.add(Word('decimal', fn = decimal))
    forth.add(Word('hex', fn = forth_hex))
    forth.add(Word('vocabulary', fn = vocabulary, runtime_fn = dovoc))
    forth.add(Word('definitions', fn = definitions))
    forth.add(Word('create', fn = create, runtime_fn = docreate))
    forth.add(Word('code', fn = code))
    forth.add(Word(':', fn = colon))
    forth.add(Word(';', fn = semi, immediate = True))
    forth.add(Word(';code', fn = semi_code, immediate = True))
    forth.add(Word('does>', fn = does, immediate = True))
    forth.add(Word("'", fn = forth_tick))
    forth.add(Word('constant', runtime_fn = run_constant))
    forth.add(Word('variable', runtime_fn = run_variable))
    forth.add(Word('user', runtime_fn = run_user))
    forth.add(Word('immediate', fn = immediate))
    forth.add(Word('+origin', fn = plus_origin))
    forth.add(Word('aux.origin', fn = aux_origin))
    forth.add(Word('[compile]', fn = br_compile, immediate = True))
    forth.add(Word('literal', fn = literal, immediate = True))
    forth.add(Word('allot', fn = allot))
    forth.add(Word('latest', fn = latest))
    forth.add(Word('."', fn = dot_quote, immediate = True))
    forth.add(Word('1+', fn = one_plus))
    forth.add(Word('2+', fn = two_plus))
    forth.add(Word('2*', fn = two_mult))
    forth.add(Word('+', fn = plus))
    forth.add(Word('-', fn = minus))
    forth.add(Word('/mod', fn = div_mod))
    forth.add(Word('swap', fn = swap))
    forth.add(Word('rot', fn = rot))
    forth.add(Word('if', fn = forth_if, immediate = True))
    forth.add(Word('else', fn = forth_else, immediate = True))
    forth.add(Word('then', fn = forth_then, immediate = True))
    forth.add(Word('do', fn = forth_do, immediate = True))
    forth.add(Word('?do', fn = query_do, immediate = True))
    forth.add(Word('loop', fn = loop, immediate = True))
    forth.add(Word('+loop', fn = plus_loop, immediate = True))
    forth.add(Word('begin', fn = begin, immediate = True))
    forth.add(Word('until', fn = until, immediate = True))
    forth.add(Word('again', fn = again, immediate = True))
    forth.add(Word('repeat', fn = repeat, immediate = True))
    forth.add(Word('while', fn = forth_while, immediate = True))
    forth.add(Word('forth', fn = dovoc, immediate = True))
    forth.add(Word('r>', fn = r_from))
    forth.add(Word('r@', fn = r_fetch))
    forth.add(Word('>r', fn = to_r))
    dicts.dicts['forth'] = forth

########################################
# ASSEMBLER dictionary words

mode = 2

def set_mode(w, m):
    global mode
    mode = m

def zaddr(w, a):
    global mode
    global stack
    mode = 3
    stack.push(a)

def cpu(w, i):
    global target
    target.ccomma(i)

asm_index = (0x09, 0x09, 0x05, 0x15, 0x15, 0x01, 0x11, 0x80,
	     0x09, 0x80, 0x0d, 0x1d, 0x19, 0x80, 0x80, 0x80,
	     0x80, 0x00, 0x04, 0x14, 0x14, 0x80, 0x80, 0x80,
	     0x80, 0x80, 0x0c, 0x1c, 0x1c, 0x80, 0x80, 0x2c)

def mcpu(w, a, b):
    global stack
    global target
    global mode

    def upmode(a, b):
        global mode
        if a:
            if mode & 8 == 0:
                mode |= 8
        return b & (1 << (mode & 0xF)) == 0

    if b & 0x80:
        mode |= 0x10
    n = stack.peek()
    if upmode(upmode(n & 0xFF00, b), b):
        global src
        src.error("Incorrect addressing mode")
    target.ccomma(asm_index[mode]+a)
    if mode & 7:
        v = stack.pop()
        if mode & 0xF < 7:
	    target.ccomma(v)
	else:
	    target.comma(v)
    mode = 2

def asm_begin(w):
    global target
    global stack
    stack.push(target.here())

def asm_until(w):
    global target
    global stack
    tst = stack.pop()
    target.ccomma(tst)
    old_here = stack.pop()
    target.ccomma(256+old_here-(target.here()+1))

def asm_if(w):
    global target
    global stack
    target.ccomma(stack.pop())
    stack.push(target.here())
    target.ccomma(0)

def asm_then(w):
    global target
    global stack
    old_here = stack.pop()
    if target.cfetch(old_here) != 0:
        target.store(old_here, target.here())
    else:
	target.cstore(old_here, target.here()-(old_here+1))

def asm_else(w):
    global target
    global stack
    old_here = stack.pop()
    new_here = target.here() + 1
    target.ccomma(0x4C)
    target.comma(1)
    target.cstore(old_here, target.here()-(old_here+1))
    stack.push(new_here)

def asm_not(w):
    global stack
    stack.push(stack.pop() + 0x20)

def label(w):
    global dicts
    global target
    name = get_word()
    dicts.dicts['assembler'].constant(name, target.here())

def tick_t(w):
    global target
    global stack
    name = get_word()
    pfa = body(name)
    stack.push(pfa)

def byte_in(w):
    global target
    global stack
    name = get_word()
    pfa = body(name)
    stack.push(stack.pop()+pfa)

def end_code(w):
    global target
    global dicts
    target.smudge()
    dicts.context = dicts.current

def setup_assembler_dict():
    global dicts
    assembler = CrossDict(dicts.dicts['forth'])
    dicts.dicts['assembler'] = assembler
    dicts.context = 'assembler'
    assembler.constant('bos', 0x10)
    assembler.constant('tos', 0x60)
    assembler.constant('n', value_get('tos')+8)
    assembler.constant('ip', value_get('n')+8)
    assembler.constant('w', value_get('ip')+3)
    assembler.constant('up', value_get('w')+2)
    assembler.constant('xsave', value_get('up')+2)
    assembler.constant('tibx', 0x100)

    assembler.add(Word('.a', set_mode, (0,)))
    assembler.add(Word('#', set_mode, (1,)))
    assembler.add(Word('mem', set_mode, (2,)))
    assembler.add(Word(',x', set_mode, (3,)))
    assembler.add(Word(',y', set_mode, (4,)))
    assembler.add(Word('x)', set_mode, (5,)))
    assembler.add(Word(')y', set_mode, (6,)))
    assembler.add(Word(')', set_mode, (0xF,)))

    assembler.add(Word('bot', zaddr, (0,)))
    assembler.add(Word('sec', zaddr, (2,)))
    assembler.add(Word('rp)', zaddr, (0x101,)))

    assembler.add(Word('brk,', cpu, (0x00,)))
    assembler.add(Word('clc,', cpu, (0x18,)))
    assembler.add(Word('cld,', cpu, (0xD8,)))
    assembler.add(Word('cli,', cpu, (0x58,)))
    assembler.add(Word('clv,', cpu, (0xB8,)))
    assembler.add(Word('dex,', cpu, (0xCA,)))
    assembler.add(Word('dey,', cpu, (0x88,)))
    assembler.add(Word('inx,', cpu, (0xE8,)))
    assembler.add(Word('iny,', cpu, (0xC8,)))
    assembler.add(Word('nop,', cpu, (0xEA,)))
    assembler.add(Word('pha,', cpu, (0x48,)))
    assembler.add(Word('php,', cpu, (0x08,)))
    assembler.add(Word('pla,', cpu, (0x68,)))
    assembler.add(Word('plp,', cpu, (0x28,)))
    assembler.add(Word('rti,', cpu, (0x40,)))
    assembler.add(Word('rts,', cpu, (0x60,)))
    assembler.add(Word('sec,', cpu, (0x38,)))
    assembler.add(Word('sed,', cpu, (0xF8,)))
    assembler.add(Word('sei,', cpu, (0x78,)))
    assembler.add(Word('tax,', cpu, (0xAA,)))
    assembler.add(Word('tay,', cpu, (0xA8,)))
    assembler.add(Word('tsx,', cpu, (0xBA,)))
    assembler.add(Word('txa,', cpu, (0x8A,)))
    assembler.add(Word('txs,', cpu, (0x9A,)))
    assembler.add(Word('tya,', cpu, (0x98,)))

    assembler.add(Word('adc,', mcpu, (0x60, 0x1C6E)))
    assembler.add(Word('and,', mcpu, (0x20, 0x1C6E)))
    assembler.add(Word('cmp,', mcpu, (0xC0, 0x1C6E)))
    assembler.add(Word('eor,', mcpu, (0x40, 0x1C6E)))
    assembler.add(Word('lda,', mcpu, (0xA0, 0x1C6E)))
    assembler.add(Word('ora,', mcpu, (0x00, 0x1C6E)))
    assembler.add(Word('sbc,', mcpu, (0xE0, 0x1C6E)))
    assembler.add(Word('sta,', mcpu, (0x80, 0x1C6C)))
    assembler.add(Word('asl,', mcpu, (0x01, 0x0D0D)))
    assembler.add(Word('dec,', mcpu, (0xC1, 0x0C0C)))
    assembler.add(Word('inc,', mcpu, (0xE1, 0x0C0C)))
    assembler.add(Word('lsr,', mcpu, (0x41, 0x0D0D)))
    assembler.add(Word('rol,', mcpu, (0x21, 0x0D0D)))
    assembler.add(Word('ror,', mcpu, (0x61, 0x0D0D)))
    assembler.add(Word('stx,', mcpu, (0x81, 0x0414)))
    assembler.add(Word('cpx,', mcpu, (0xE0, 0x0486)))
    assembler.add(Word('cpy,', mcpu, (0xC0, 0x0486)))
    assembler.add(Word('ldx,', mcpu, (0xA2, 0x1496)))
    assembler.add(Word('ldy,', mcpu, (0xA0, 0x0C8E)))
    assembler.add(Word('sty,', mcpu, (0x80, 0x048C)))
    assembler.add(Word('jsr,', mcpu, (0x14, 0x0480)))
    assembler.add(Word('jmp,', mcpu, (0x40, 0x8480)))
    assembler.add(Word('bit,', mcpu, (0x20, 0x0484)))

    assembler.add(Word('begin,', asm_begin))
    assembler.add(Word('until,', asm_until))
    assembler.add(Word('if,', asm_if))
    assembler.add(Word('then,', asm_then))
    assembler.add(Word('else,', asm_else))
    assembler.add(Word('not', asm_not))

    assembler.constant('cs', 0x90)
    assembler.constant('0=', 0xD0)
    assembler.constant('0<', 0x10)
    assembler.constant('>=', 0x90)
    assembler.constant('vs', 0x50)

    assembler.add(Word('label', label))
    assembler.add(Word("'t", tick_t, immediate = True))
    assembler.add(Word('byte.in', byte_in))
    assembler.add(Word('end-code', end_code))

    dicts.dicts['forth'].add(Word('assembler', fn = dovoc, immediate = True))
    dicts.context = 'forth'

dicts = DictChain()
stack = Stack()
rstack = Stack()
xstack = Stack()
src = None
target = None
options = None

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-s", "--start", dest="startaddr", type="int",
                      help="Main start address", metavar="addr")
    parser.add_option("-a", "--aux", dest="auxaddr", type="int",
                      help="Auxiliary start address", metavar="addr")
    parser.add_option("-o", "--output", dest="outfile", type="str",
                      help="Output file name", metavar="outf")
    parser.add_option("-D", "--dump", dest="dump",
                      action="store_true", default=False)
    parser.add_option("-d", "--debug", dest="debug", action="count")
    (options, args) = parser.parse_args()

    if options.startaddr is None:
        parser.error("Missing start address");
    if options.outfile is None:
        parser.error("Missing output filename");
    if len(args) == 0:
        parser.error("Missing input filename");

    setup_forth_dict()
    setup_assembler_dict()
    dicts.current = 'forth'
    dicts.context = 'forth'

    target = TargetDict(options.startaddr, options.auxaddr)

    src = SourceFile(args[0])
    tok = True
    while tok:
        tok = src.token(' ')
        if tok:
            interpret(tok.lower())

    if options.debug:
        print "Fixing up forward references"
    for k, v in target.fwd_refs.items():
        dicts.current, dicts.context, name = k
        if options.debug > 1:
            print "   ", dicts.current, dicts.context, name, len(v)
        w = find(name)
        if w is None:
            raise RuntimeError("Forward reference to undefined word "+name)
        for addr in v:
            target.store(addr, target.tick(w))

    if options.dump:
        print "Data stack: ", stack
        print "Dictionaries: ", len(dicts.dicts.keys())
        print "Memory region start: 0x%04x" % target.main_block.start
        print target.main_block
        if target.aux_block:
            print "Auxiliary region start: 0x%04x" % target.aux_block.start
            print target.aux_block

    if options.debug:
        print "Writing binary to", options.outfile
    if options.debug > 1:
        print "    Memory region length: 0x%04x" % len(target.main_block.mem)
    fh = open(options.outfile, 'w')
    fh.write(target.main_block.mem)
    if target.aux_block:
        fh.write(target.aux_block.mem)
    fh.close()
